
#include <windows.h>

#include "wnaspi32.h"
#include "scsidefs.h"
#include "nnnaspi.h"


static HINSTANCE WinASPI32DLL = NULL;
static int nHostAdapter = 0;
static struct nnn_cddrive_info *drivelist=NULL;
static int ncdromdevice=0;
static struct TOC *pDisc_TOC=NULL;


static DWORD (*pGetASPI32SupportInfo)(void);
static DWORD (*pSendASPI32Command)(LPSRB);



int InitASPIDLL(){
	int Status;

	if(WinASPI32DLL != NULL)
		return 1;
	WinASPI32DLL = LoadLibrary("WNASPI32");
	if(WinASPI32DLL != NULL){
		pGetASPI32SupportInfo = (DWORD (*)(void))GetProcAddress(WinASPI32DLL, "GetASPI32SupportInfo");
		pSendASPI32Command = (DWORD (*)(LPSRB))GetProcAddress(WinASPI32DLL, "SendASPI32Command");
		if(pGetASPI32SupportInfo == NULL || pSendASPI32Command == NULL){
			ReleaseASPIDLL();
			return 0;
		}
		Status = pGetASPI32SupportInfo();
		nHostAdapter = Status&0xff;
		Status = (Status>>8)&0xff;
		if(SS_NO_ADAPTERS == Status || SS_COMP != Status){
			ReleaseASPIDLL();
			return 0;
		}
		pDisc_TOC = (struct TOC *)malloc(sizeof(struct TOC));
		return 1;
	}
	return 0;
}


void ReleaseASPIDLL(){
	FreeCDROMDeviceInfo();
	if(WinASPI32DLL != NULL){
		FreeLibrary(WinASPI32DLL);
		WinASPI32DLL=NULL;
		nHostAdapter = 0;
	}
}


// CD-ROMhCu
void EnumCDROMDevice(){
	int nAdapterMax, nDevice, i, j;
	SRB_HAInquiry HAInquiry;
	SRB_GDEVBlock GDEVBlock;

	if(WinASPI32DLL == NULL)
		return;
	nAdapterMax = nHostAdapter;
	for(i=0; i<nAdapterMax; i++){

		memset( &HAInquiry, 0, sizeof(SRB_HAInquiry) );
		HAInquiry.SRB_Cmd = SC_HA_INQUIRY;
		HAInquiry.SRB_HaId = i;
		pSendASPI32Command(&HAInquiry);
		if(SS_COMP != HAInquiry.SRB_Status){
			continue;
		}
		nDevice = HAInquiry.HA_Unique[3];
		for(j=0; j<nDevice; j++){
			if(j == HAInquiry.HA_SCSI_ID)
				continue;
			memset(&GDEVBlock, 0, sizeof(GDEVBlock));
			GDEVBlock.SRB_Cmd		= SC_GET_DEV_TYPE;
			GDEVBlock.SRB_HaId		= i;				// SCSIzXgA_v^ԍ
			GDEVBlock.SRB_Target	= j;				// foCXԍ
//			GDEVBlock.SRB_Lun		= 0;				// WJjbgԍ

			pSendASPI32Command((LPSRB)&GDEVBlock);
			if(SS_COMP != GDEVBlock.SRB_Status){
				continue;
			}
			if(DTYPE_CROM == GDEVBlock.SRB_DeviceType){
				struct nnn_cddrive_info *tmp;
				int an = ncdromdevice;
				an++;
				tmp = (struct nnn_cddrive_info *)realloc(drivelist, sizeof(struct nnn_cddrive_info) * an);
				if(tmp==NULL)
					continue;
				drivelist = tmp;
				drivelist[ncdromdevice].HostID = i;			// HostAdapter No
				drivelist[ncdromdevice].DeviceID=j;			// Device No
				strcpy(drivelist[ncdromdevice].HostAdapterName , (const char *)HAInquiry.HA_Identifier);	// HostAdapter name
				ASPI_GetDriveInfo(&drivelist[ncdromdevice]);
				ncdromdevice++;
			}
		}
	}
}


void FreeCDROMDeviceInfo(){
	if(pDisc_TOC){
		free(pDisc_TOC);
		pDisc_TOC = NULL;
	}
	if(drivelist != NULL){
		free(drivelist);
		drivelist = NULL;
	}
	ncdromdevice=0;
}


// foCX擾
int ASPI_GetDriveInfo(struct nnn_cddrive_info *driveinfo){

	SRB_ExecSCSICmd ExecSCSICmd;
	DWORD dwStatus;
	BYTE reply_buf[255];					// SCSIR}h̉pobt@
	HANDLE hEvent;							// CxgIuWFNg̃nh

	if(WinASPI32DLL == NULL)
		return 0;
	hEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
	if(NULL == hEvent){
		return 0;
	}

	memset(reply_buf, 0, sizeof(reply_buf));

	memset(&ExecSCSICmd, 0, sizeof(ExecSCSICmd));
	ExecSCSICmd.SRB_Cmd		= SC_EXEC_SCSI_CMD;		// R}h
	ExecSCSICmd.SRB_HaId	= driveinfo->HostID;	// SCSIzXgA_v^ԍ
	ExecSCSICmd.SRB_Flags	= SRB_DIR_IN | SRB_EVENT_NOTIFY;
	ExecSCSICmd.SRB_Target	= driveinfo->DeviceID;	// SCSI ID
	ExecSCSICmd.SRB_Lun		= 0;					// WJjbgԍ
	ExecSCSICmd.SRB_BufLen		= sizeof(reply_buf);	// obt@TCY
	ExecSCSICmd.SRB_BufPointer	= reply_buf;			// obt@̃|C^
	ExecSCSICmd.SRB_SenseLen	= SENSE_LEN;			// ZXobt@̃TCY
	ExecSCSICmd.SRB_CDBLen		= 6;					// CDBBytẽTCY
	ExecSCSICmd.SRB_PostProc	= hEvent;				// CxgIuWFNg
	ExecSCSICmd.CDBByte[0]		= SCSI_INQUIRY;			// SCSIR}h̎
	ExecSCSICmd.CDBByte[4]		= sizeof(reply_buf);	// obt@TCY

	ResetEvent(hEvent);

	dwStatus = pSendASPI32Command((LPSRB)&ExecSCSICmd);
	if(SS_PENDING == dwStatus){
		WaitForSingleObject(hEvent, INFINITE);
	}
	CloseHandle(hEvent);

	if(SS_COMP != ExecSCSICmd.SRB_Status){
		return 0;
	}

	// foCXRs[
	memcpy(driveinfo->VendorIDName, &reply_buf[8], 8);
	memcpy(driveinfo->ProductIDName, &reply_buf[16], 16);
	memcpy(driveinfo->ProductVerName, &reply_buf[32], 4);
	return 1;
}


// fBAZbgĂ邩`FbN
int ASPI_CheckMediaReady(struct nnn_cddrive_info *driveinfo){

	SRB_ExecSCSICmd ExecSCSICmd;
	DWORD dwStatus;
	HANDLE hEvent;							// CxgIuWFNg̃nh

	if(WinASPI32DLL == NULL)
		return 0;
	hEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
	if(NULL == hEvent){
		return 0;
	}

	memset(&ExecSCSICmd, 0, sizeof(ExecSCSICmd));
	ExecSCSICmd.SRB_Cmd		= SC_EXEC_SCSI_CMD;		// R}h
	ExecSCSICmd.SRB_HaId	= driveinfo->HostID;	// SCSIzXgA_v^ԍ
	ExecSCSICmd.SRB_Flags	= SRB_DIR_SCSI | SRB_EVENT_NOTIFY;
	ExecSCSICmd.SRB_Target	= driveinfo->DeviceID;	// SCSI ID
	ExecSCSICmd.SRB_Lun		= 0;					// WJjbgԍ
	ExecSCSICmd.SRB_SenseLen	= SENSE_LEN;			// ZXobt@̃TCY
	ExecSCSICmd.SRB_CDBLen		= 6;					// CDBBytẽTCY
	ExecSCSICmd.SRB_PostProc	= hEvent;				// CxgIuWFNg
	ExecSCSICmd.CDBByte[0]		= SCSI_TST_U_RDY;		// SCSIR}h̎

	ResetEvent(hEvent);

	dwStatus = pSendASPI32Command((LPSRB)&ExecSCSICmd);
	if(SS_PENDING == dwStatus){
		WaitForSingleObject(hEvent, INFINITE);
	}
	CloseHandle(hEvent);

	if(SS_COMP == ExecSCSICmd.SRB_Status){
		return 1;
	}
	return 0;
}


// read TOC
int ASPI_ReadTOC(struct nnn_cddrive_info *driveinfo){

	SRB_ExecSCSICmd ExecSCSICmd;
	DWORD dwStatus;
	HANDLE hEvent;							// CxgIuWFNg̃nh

	if(WinASPI32DLL == NULL)
		return 0;
	if(pDisc_TOC == NULL)
		return 0;
	memset(&pDisc_TOC, 0, sizeof(struct TOC));
	hEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
	if(NULL == hEvent){
		return 0;
	}

	memset(&ExecSCSICmd, 0, sizeof(ExecSCSICmd));
	ExecSCSICmd.SRB_Cmd		= SC_EXEC_SCSI_CMD;		// R}h
	ExecSCSICmd.SRB_HaId	= driveinfo->HostID;	// SCSIzXgA_v^ԍ
	ExecSCSICmd.SRB_Target	= driveinfo->DeviceID;	// SCSI ID
	ExecSCSICmd.SRB_Lun		= 0;					// WJjbgԍ
	ExecSCSICmd.SRB_Flags	= SRB_DIR_IN | SRB_EVENT_NOTIFY;
	ExecSCSICmd.SRB_BufLen	= sizeof(struct TOC);
	ExecSCSICmd.SRB_BufLen	= sizeof(struct TOC);
	ExecSCSICmd.SRB_BufPointer	= (BYTE *)pDisc_TOC;
	ExecSCSICmd.SRB_SenseLen	= SENSE_LEN;	/*0x0e*/// ZXobt@̃TCY
	ExecSCSICmd.SRB_CDBLen		= 0x0a;			/*6*/	// CDBBytẽTCY
	ExecSCSICmd.SRB_PostProc	= hEvent;				// CxgIuWFNg
	ExecSCSICmd.CDBByte[0]		= 0x43;					// SCSIR}h̎
	ExecSCSICmd.CDBByte[7]		= 0x03;		// ofs. 7-8 used for toc len
	ExecSCSICmd.CDBByte[8]		= 0x24;		// TOC buffer length == 0x324

	ResetEvent(hEvent);

	dwStatus = pSendASPI32Command((LPSRB)&ExecSCSICmd);
	if(SS_PENDING == dwStatus){
		WaitForSingleObject(hEvent, INFINITE);
	}
	CloseHandle(hEvent);
	return 1;
}





// Read Raw Data
int ASPI_ReadRawData(struct nnn_cddrive_info *driveinfo, int Sector, unsigned char *pMem){

	SRB_ExecSCSICmd ExecSCSICmd;
	DWORD dwStatus;
	HANDLE hEvent;							// CxgIuWFNg̃nh
	unsigned char *pBuffer=NULL;

	if(WinASPI32DLL == NULL)
		return 0;
	pBuffer = (unsigned char *)malloc(2352);
	if(pBuffer == NULL)
		return 0;

	hEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
	if(NULL == hEvent){
		return 0;
	}

	memset(&ExecSCSICmd, 0, sizeof(ExecSCSICmd));
	ExecSCSICmd.SRB_Cmd		= SC_EXEC_SCSI_CMD;		// R}h
	ExecSCSICmd.SRB_HaId	= driveinfo->HostID;	// SCSIzXgA_v^ԍ
	ExecSCSICmd.SRB_Flags	= SRB_DIR_IN | SRB_EVENT_NOTIFY;
	ExecSCSICmd.SRB_Target	= driveinfo->DeviceID;	// SCSI ID
	ExecSCSICmd.SRB_Lun		= 0;					// WJjbgԍ
	ExecSCSICmd.SRB_BufLen = 2352;					// obt@̃TCY
	ExecSCSICmd.SRB_BufPointer = pBuffer;			// obt@
	ExecSCSICmd.SRB_SenseLen	= SENSE_LEN;			// ZXobt@̃TCY
	ExecSCSICmd.SRB_CDBLen		= 12;					// CDBBytẽTCY
	ExecSCSICmd.SRB_PostProc	= hEvent;				// CxgIuWFNg
	ExecSCSICmd.CDBByte[0]		= 0xBE;	/* read CD Raw */	// SCSIR}h̎
	ExecSCSICmd.CDBByte[3]		= (Sector >> 16) & 0xFF;
	ExecSCSICmd.CDBByte[4]		= (Sector >> 8) & 0xFF;
	ExecSCSICmd.CDBByte[5]		= Sector & 0xFF;
	ExecSCSICmd.CDBByte[8]		= 1;
	ExecSCSICmd.CDBByte[8]		= 0xF0;

	ResetEvent(hEvent);

	dwStatus = pSendASPI32Command((LPSRB)&ExecSCSICmd);
	if(SS_PENDING == dwStatus){
		WaitForSingleObject(hEvent, INFINITE);
	}
	CloseHandle(hEvent);

	memcpy(pMem, pBuffer+0x10, 2048);
	free(pBuffer);
	if(SS_COMP == ExecSCSICmd.SRB_Status){
		return 1;
	}
	return 0;
}

